/*
*	Proxy HTTP daemon with automatic charset detection
*
*			Cyrillic Wizard
*
*	Copyright (C) 1996 Alexey V. Kuznetsov (avk@gamma.ru), Gamma Ltd.
*
*/

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <errno.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>
#include <syslog.h>
#include <time.h>
#include <fcntl.h>
#include <netdb.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#define PROG_NAME		"CyrWiz"
#define VERSION			"0.2"
#define DEFAULT_CONF_FILE	"/usr/local/lib/cyrwiz/cyrwiz.conf"

char *tab_dir=0;	/* directory with code tables */
char *msg_dir=0;	/* directory with error messages */
char *def_host=0;	/* default host */
char *mst_prox=0;	/* master proxy */
char *log_file=0;	/* log file name */
char *char_arg=0;	/* charset from command line */
char *our_addr=0;	/* our address */
char  sim_prox=0;	/* simple or http proxy */

struct MultString {
	char *string;
	struct MultString *next;
};

struct MultString *allowed_hosts=0;
struct MultString *deny_hosts=0;

static char *wdays[7] = {
	"Sun",
	"Mon",
	"Tue",
	"Wed",
	"Thu",
	"Fri",
	"Sat"
};

static char *months[12] = {
	"Jan",
	"Feb",
	"Mar",
	"Apr",
	"May",
	"Jun",
	"Jul",
	"Aug",
	"Sep",
	"Oct",
	"Nov",
	"Dec"
};

char *http_date ( void )
{
	time_t t_loc;
	struct tm *tm;
	static char d_str[] = { "Mon, 01 Jan 1970 00:00:00 GMT\0" };
	t_loc = time ( NULL );
	tm = gmtime ( &t_loc );
	snprintf( d_str,sizeof d_str,"%3s, %02d %3s %04d %02d:%02d:%02d GMT",wdays[tm->tm_wday],tm->tm_mday,months[tm->tm_mon],tm->tm_year+1900,tm->tm_hour,tm->tm_min,tm->tm_sec );
	return d_str;
}

char *c_date ( void )
{
	time_t t_loc;
	struct tm *tm;
	static char d_str[10];
	t_loc = time ( NULL );
	tm = localtime ( &t_loc );
	snprintf( d_str,sizeof(d_str),"%02d.%02d.%02d",tm->tm_mday,tm->tm_mon+1,tm->tm_year );
	return d_str;
}

char *c_time ( void )
{
	time_t t_loc;
	struct tm *tm;
	static char t_str[10];
	t_loc = time ( NULL );
	tm = localtime ( &t_loc );
	snprintf( t_str,sizeof(t_str),"%02d:%02d:%02d",tm->tm_hour,tm->tm_min,tm->tm_sec );
	return t_str;
}

#define MAX_LOG_LINE	1024

char log_line[MAX_LOG_LINE+1];
char *end_head;
char log_open=0;

#define	N_PRIORS	3
char priors[N_PRIORS] = { 'I','W','E' };

#define	L_INFO	0
#define	L_WARN	1
#define	L_ERR	2

void m_log ( int pri, char *fmt, ... )
{
	va_list ap;
	static char pri_str[3] = { "I " };
	if( !log_open ) return;
	if ( pri >= N_PRIORS ) pri = 2;
	pri_str[0] = priors[pri];
	va_start( ap,fmt );
	strcpy( end_head,pri_str );
	strcat( end_head,c_time() );
	strcat( end_head," " );
	strcat( end_head,fmt );
	strcat( end_head,"\n" );
	vfprintf ( stderr,log_line,ap );
	fflush( stderr );
}

void m_perror ( char *name )
{
	if( !log_open ) return;
	strcpy( end_head,"E " );
	strcat( end_head,c_time() );
	strcat( end_head," " );
	strcat( end_head,name );
	strcat( end_head," : %s\n" );
	fprintf( stderr,log_line,strerror(errno) );
}

void m_openlog ( char *name, char *file )
{
	sprintf( log_line,"%s[%d]: ",name,(int)getpid() );
	end_head = log_line+strlen(log_line);
	log_open = 1;
	if ( file == NULL ) return;
	if ( freopen ( file,"a",stderr ) == NULL ) {
		m_log( L_ERR,"can not open %s : %s",file,strerror(errno) );
	}
}

void *c_malloc( int size )
{
	void *ptr = malloc(size);
	if ( ptr == NULL ) {
		m_perror( "malloc" );
		syslog( LOG_ERR,"malloc : %m" );
		exit( 1 );
	}
	return ptr;
}

void strdcpy( char **to, char *from )
{
	int i = strlen( from );
	*to = c_malloc( i+1 );
	memcpy( *to,from,i+1 );
}

/*
	subroutine for reading string ending with LF or CRLF
*/

#define MAX_STRING	4096
char inp_string[MAX_STRING+1];

char *get_string ( int fd )
{
	char *ptr = inp_string;
	int counter = 0;
	int rc;
	for ( ;; ) {
		if ( (rc = read ( fd,ptr,1 )) == (-1) ) return 0;
		if ( rc == 0 ) {
			if ( ptr == inp_string ) return 0;
			*ptr = 0;
			return inp_string;
		}
		if ( *ptr == '\n' ) {
			*ptr = 0;
			return inp_string;
		}
		if ( *ptr == '\r' ) continue;
		if ( counter < MAX_STRING ) {
			counter++; ptr++;
		}
	}
}

/*
	read config file
*/

void read_arg ( char **param )
{
	char *token = strtok( NULL," \t" );
	if ( token == NULL ) {
		syslog( LOG_ERR,"argument in config file not set" );
		return;
	}
	strdcpy( param,token );
}

void read_barg ( char *param )
{
	char *token = strtok( NULL," \t" );
	if ( token == NULL ) {
		syslog( LOG_ERR,"argument in config file not set" );
		return;
	}
	if( *token == 'Y' || *token == 'y' || *token == 'T' || *token == 't' ) {
		*param = 1;
	} else {
		*param = 0;
	}
}

void read_marg ( struct MultString **param )
{
	char *token = strtok( NULL," \t" );
	struct MultString *ms1;
	if ( token == NULL ) {
		syslog( LOG_ERR,"argument in config file not set" );
		return;
	}
	ms1 = c_malloc( sizeof( struct MultString ) );
	strdcpy( &ms1->string,token );
	ms1->next = *param;
	*param = ms1;
}

typedef (*p_func)(void**);

struct ConfParam {
	char *name;
	void **param;
	p_func func;
};

struct ConfParam params[] = {
	{ "DefaultHost", (void**)&def_host,     (p_func)read_arg },	/* default host */
	{ "MasterProxy", (void**)&mst_prox,     (p_func)read_arg },	/* default host */
	{ "Tables",      (void**)&tab_dir,      (p_func)read_arg },	/* code tables directory */
	{ "Messages",    (void**)&msg_dir,      (p_func)read_arg },	/* messages directory */
	{ "LogFile",     (void**)&log_file,     (p_func)read_arg },	/* log file */
	{ "OurAddress",  (void**)&our_addr,     (p_func)read_arg },	/* log file */
	{ "AllowedHosts",(void**)&allowed_hosts,(p_func)read_marg },	/* allowed hosts */
	{ "DenyHosts",   (void**)&deny_hosts,   (p_func)read_marg },	/* deny hosts */
	{ "SimpleProxy", (void**)&sim_prox,     (p_func)read_barg },	/* deny hosts */
	{ 0,            0,        0        }
};

int load_conf ( char *conf_file )
{
	int fd = open ( conf_file,O_RDONLY,0 );
	char *str;
	int i;
	if ( fd == (-1) ) {
		syslog( LOG_ERR,"Can not open config file %s : %s",conf_file,strerror(errno) );
		return 1;
	}
	while ( (str = get_string ( fd )) ) {
		char *token;
		if ( *str == 0 || *str == '#' ) continue;
		token = strtok ( str," \t" );
		for ( i=0; params[i].name; i++ ) {
			if ( strcmp( params[i].name,token ) ) continue;
			params[i].func( params[i].param );
			goto param_found;
		}
		syslog( LOG_ERR,"parameter '%s' not found",token );
		param_found:;
	}
	return 0;
}

int match_string ( char *string, char *pattern )
{
	char *aster = strchr ( pattern,'*' );
	int i;
	if ( aster == NULL ) {
		return strcasecmp( string,pattern );
	}
	if ( aster != pattern ) {
		if ( strncasecmp( string,pattern,aster-pattern ) ) return 1;
	}
	aster++;
	i = strlen( aster );
	if ( i == 0 ) return 0;
	return strcasecmp( string+strlen(string)-i,aster );
}

int is_bad_host ( char *host )
{
	struct MultString *ms;
	for ( ms=deny_hosts; ms; ms=ms->next ) {
		if ( match_string( host,ms->string ) == NULL ) return 1;
	}
	for ( ms=allowed_hosts; ms; ms=ms->next ) {
		if ( match_string( host,ms->string ) == NULL ) return 0;
	}
	return 1;
}

/*
	load to memory header of message
*/

struct Header {
	char *name;
	char *value;
	struct Header *next;
};

struct Header *load_header ( int fd )
{
	char *line;
	struct Header *header=0,*cur_head=0;
	while ( (line = get_string( fd )) ) {
		char *ptr=0;
		if ( *line ) {
			ptr = strchr(line,':');
			if ( ptr == NULL ) {
				m_log( L_ERR,"invalid line in header : %s",line );
				return 0;
			}
		}
		if ( header == NULL ) {
			header = c_malloc( sizeof( struct Header ) );
			header->next = 0; cur_head = header;
		} else {
			cur_head->next = c_malloc( sizeof( struct Header ) );
			cur_head = cur_head->next;
			cur_head->next = 0;
		}
		if ( *line == 0 ) {
			cur_head->name = cur_head->value = "\0";
			break;
		}
		cur_head->name = c_malloc( ptr-line+1 );
		memcpy( cur_head->name,line,ptr-line );
		cur_head->name[ptr-line] = 0;
		ptr++;
		while ( (*ptr == ' ' || *ptr == '\t') && *ptr ) ptr++;
		if ( *ptr ) {
			strdcpy( &cur_head->value,ptr );
		} else {
			cur_head->value = "\0";
		}
	}
	return header;
}

void dump_header ( int fd, struct Header *header )
{
	for( ; header; header = header->next ) {
		if ( *header->name == 0 ) {
			write( fd,"\r\n",2 );
			break;
		}
		strcpy( inp_string,header->name );
		strcat( inp_string,": " );
		strcat( inp_string,header->value );
		strcat( inp_string,"\r\n" );
		if ( write( fd,inp_string,strlen( inp_string ) ) == (-1) ) {
			m_perror( "write" );
		}
	}
}

void log_header ( struct Header *header )
{
	for( ; header; header = header->next ) {
		if ( *header->name == 0 ) {
			break;
		}
		m_log( L_INFO,"%s: %s",header->name,header->value );
	}
}

char *look_header ( struct Header *header, char *name )
{
	for ( ; header; header=header->next ) {
		if ( strcasecmp ( name,header->name ) == 0 ) {
			return header->value;
		}
	}
	return 0;
}

int set_header ( struct Header *header, char *name, char *value )
{
	for ( ; header; header=header->next ) {
		if ( strcasecmp ( name,header->name ) == 0 ) {
			free ( header->value );
			header->value = value;
			return 0;
		}
	}
	return 1;
}

int add_header ( struct Header **header, char *name, char *value )
{
	struct Header *h=*header,*h1;
	for ( ; h && h->next; h=h->next );
	h1 = c_malloc( sizeof( struct Header ) );
	h1->next = 0;
	h1->name = name;
	h1->value = value;
	if ( h ) {
		h->next = h1;
	} else {
		*header = h1;
	}
	return 0;
}

struct Method {
	char *method;
	char *uri;
	struct {
		int major;
		int minor;
	} version;
};

struct Method *load_method ( int fd )
{
	char *line,*token;
	struct Method *meth;
	line = get_string ( fd );
	token = strtok ( line," \t" );
	if ( token == NULL ) return 0;
	meth = c_malloc( sizeof( struct Method ) );
	strdcpy( &meth->method,token );
	token = strtok ( 0," \t" );
	if ( token == NULL ) return 0;
	meth->uri = c_malloc( strlen(token)+2 );
	strcpy( meth->uri,token );
	token = strtok ( 0," \t" );
	if ( token == NULL ) {
		meth->version.major = 0;
		meth->version.minor = 9;
	} else {
		if ( strncmp( token,"HTTP/",5 ) ) return 0;
		token += 5;
		token = strtok( token,"." );
		if ( token == NULL ) return 0;
		meth->version.major = atoi( token );
		token = strtok( 0," \t" );
		if ( token == NULL ) return 0;
		meth->version.minor = atoi( token );
	}
	return meth;
}

void dump_method ( int fd, struct Method *meth )
{
	if ( meth == NULL ) return;
	if ( meth->version.major < 1 ) {
		sprintf( inp_string,"%s %s\r\n",meth->method,meth->uri );
	} else {
		sprintf( inp_string,"%s %s HTTP/%d.%d\r\n",meth->method,meth->uri,meth->version.major,meth->version.minor );
	}
	if ( write ( fd,inp_string,strlen(inp_string) ) == (-1) ) {
		m_perror( "write" );
	}
}

void log_method ( struct Method *meth )
{
	if ( meth == NULL ) return;
	if ( meth->version.major < 1 ) {
		m_log( L_INFO,"%s %s",meth->method,meth->uri );
	} else {
		m_log( L_INFO,"%s %s HTTP/%d.%d",meth->method,meth->uri,meth->version.major,meth->version.minor );
	}
}

struct Answer {
	struct {
		int major;
		int minor;
	} version;
	int ret_code;
	int h_ret_code;
	char *answer;
};

struct Answer *load_answer ( int fd )
{
	char *line,*token,*ptr;
	struct Answer *answer;
	line = get_string( fd );
	token = strtok( line," \t" );
	if ( token == NULL ) return 0;
	if ( strncmp( token,"HTTP/",5 ) ) return 0;
	token += 5;
	if ( (ptr = strchr( token,'.' )) == NULL ) return 0;
	*ptr = 0;
	answer = c_malloc( sizeof( struct Answer ) );
	answer->version.major = atoi( token ); ptr++;
	answer->version.minor = atoi( ptr );
	token = strtok( 0," \t" );
	if ( token == NULL ) return 0;
	answer->ret_code = atoi( token );
	answer->h_ret_code = answer->ret_code/100;
	token += strlen( token ) + 1;
	while ( *token && (*token == ' ' || *token == '\t') );
	if ( *token ) {
		strdcpy( &answer->answer,token );
	} else {
		answer->answer = "\0";
	}
	return answer;
}

void dump_answer ( int fd, struct Answer *answer )
{
	sprintf( inp_string,"HTTP/%d.%d %d %s\r\n",answer->version.major,answer->version.minor,answer->ret_code,answer->answer );
	write( fd,inp_string,strlen(inp_string) );
}

void log_answer ( struct Answer *answer )
{
	m_log( L_INFO,"HTTP/%d.%d %d %s",answer->version.major,answer->version.minor,answer->ret_code,answer->answer );
}

/*
	subroutines for name resolution
*/

struct in_addr getaddr ( char *addr )
{
  struct in_addr ipaddr;
  if ( isdigit ( *addr ) ) {
    ipaddr.s_addr = inet_addr ( addr );
  } else {
    struct hostent *hst;
    hst = gethostbyname ( addr );
    if ( hst == 0 ) {
      ipaddr.s_addr = 0;
    } else {
      ipaddr.s_addr = *(u_long*)hst->h_addr;
    }
  }
  return ipaddr;
}

char* gethname ( struct in_addr addr )
{
  struct hostent *hst;
  char *name,*ptr;
  hst = gethostbyaddr ( (char*)&addr,4,AF_INET );
  if ( hst == 0 ) {
    ptr = inet_ntoa ( addr );
  } else {
    ptr = hst->h_name;
  }
  strdcpy( &name,ptr );
  return name;
}

/*
	subroutines for coding
*/

char *cp1251_aliases[] = {
	"x-cp1251",
	"cp1251",
	"x-cp-1251",
	"cp-1251",
	"x-windows-1251",
	"windows-1251",
	"x-win-1251",
	"win-1251",
	"x-windows1251",
	"windows1251",
	"x-win1251",
	"win1251",
	0
};

char *iso_aliases[] = {
	"iso-8859-5",
	"iso-8859_5",
	"iso_8859-5",
	"iso_8859_5",
	"iso8859-5",
	"iso8859_5",
	0
};

char *koi_aliases[] = {
	"koi8-r",
	"koi8_r",
	"koi8r",
	"koi8",
	0
};

char *cp866_aliases[] = {
	"x-cp866",
	"cp866",
	"x-cp-866",
	"cp-866",
	"x-ibm-866",
	"ibm-866",
	"x-ibm866",
	"ibm866",
	0
};

char *mac_aliases[] = {
	"x-maccyr",
	"x-mac-cyr",
	"x-ru-mac",
	"x-rumac",
	"mac-cyr",
	"maccyr",
	"ru-mac",
	"rumac",
	0
};

struct {
	char **aliases;
	char *name;
} SupCharsets[] = {
	{ koi_aliases,		"koi8-r"     },
	{ iso_aliases,		"iso-8859-5" },
	{ cp1251_aliases,	"x-cp1251"   },
	{ cp866_aliases,	"x-cp866"    },
	{ mac_aliases,		"x-maccyr"   },
	{ NULL,			NULL	     }
};

int sup_charset ( char **cset )
{
	int i;
	for ( i=0; SupCharsets[i].aliases; i++ ) {
		int j;
		char **c_aliases;
		if ( strcasecmp( SupCharsets[i].name,*cset ) == 0 ) {
			strcpy( *cset,SupCharsets[i].name );
			return 0;
		}
		for( j=0,c_aliases=SupCharsets[i].aliases; c_aliases[j]; j++ ) {
			if( strcasecmp( c_aliases[j],*cset ) == 0 ) {
				free( *cset );
				strdcpy( cset,SupCharsets[i].name );
				return 0;
			}
		}
	}
	return 1;
}

int do_recode=0;
char *charset=0,*client_charset=0;

int LoadTable ( char *table,char *charset1,char *charset2 )
{
	char *tab_path;
	int fd;
	tab_path = c_malloc( strlen(tab_dir)+strlen(charset1)+strlen(charset2)+1+1+1 );
	strcpy( tab_path,tab_dir );
	strcat( tab_path,"/" );
	strcat( tab_path,charset1 );
	strcat( tab_path,"." );
	strcat( tab_path,charset2 );
	if ( (fd = open( tab_path,O_RDONLY,0 )) == (-1) ) {
		m_log( L_ERR,"can not open %s : %s",tab_path,strerror(errno) );
		do_recode = 0;
		return 1;
	}
	if ( read( fd,table,256 ) != 256 ) {
		m_log( L_ERR,"invalid code table %s",tab_path );
		do_recode = 0;
		close ( fd );
		return 1;
	}
	close ( fd );
	free ( tab_path );
	return 0;
}

int LoadCodeTable ( char *table,char *charset1,char *charset2 )
{
	if ( strcmp( charset1,"koi8-r" ) == 0 ||
	     strcmp( charset2,"koi8-r" ) == 0 ) {
		if ( LoadTable ( table,charset1,charset2 ) ) return 1;
	} else {
		int i;
		unsigned char to_koi[256],from_koi[256];
		if ( LoadTable ( to_koi,charset1,"koi8-r" ) ) return 1;
		if ( LoadTable ( from_koi,"koi8-r",charset2 ) ) return 1;
		for ( i = 0; i < 256; i++ ) {
			table[i] = from_koi[to_koi[i]];
		}
	}
	return 0;
}

void InitRecode ( struct Header *header,char *u_charset )
{
	char *content_type = look_header( header,"Content-type" );
	char *ptr;
	char *new_value;
	do_recode = 0; charset = 0;
	if ( content_type == NULL || strncmp( content_type,"text/",5 ) ) {
		return;
	}
	if ( look_header( header,"Content-encoding" ) ) return;
	do_recode = 1;
	client_charset = u_charset;
	ptr = content_type+5;
	for ( ;; ) {
		char *c_name;
		while ( *ptr != ';' && *ptr ) ptr++;
		if ( *ptr == 0 ) break;
		ptr++;
		while ( (*ptr == ' ' || *ptr == '\t') && *ptr ) ptr++;
		if ( *ptr == 0 ) break;
		if ( strncasecmp( ptr,"charset=",8 ) ) continue;
		ptr += 8;
		if ( *ptr == '\"' ) ptr++;
		c_name = ptr;
		while ( *ptr != '\"' && *ptr != ';' && *ptr ) ptr++;
		if ( ptr == c_name ) break;
		charset = c_malloc( ptr-c_name+1 );
		strncpy( charset,c_name,ptr-c_name );
		if ( sup_charset( &charset ) ) {
			do_recode = 0;
			return;
		}
		if ( u_charset == NULL ) return;
		new_value = c_malloc( strlen( content_type )-(ptr-c_name)+strlen( u_charset )+1 );
		strncpy( new_value,content_type,c_name-content_type );
		strcat( new_value,u_charset );
		strcat( new_value,ptr );
		set_header( header,"Content-type",new_value );
		return;
	}
	if ( u_charset == NULL ) return;
	new_value = c_malloc( strlen( content_type )+13 );
	strcpy( new_value,content_type );
	strcat( new_value,"; charset=\"" );
	strcat( new_value,u_charset );
	strcat( new_value,"\"" );
	set_header( header,"Content-type",new_value );
}

void ResetRecode ( void )
{
	do_recode = 0; charset = 0;
}

void code_buf( unsigned char *table,unsigned char *buf,int buflen )
{
	while( buflen ) {
		*buf = table[*buf];
		buf++; buflen--;
	}
}

unsigned char code_table[256];

void RecodeBuffer ( unsigned char *buf, int buflen )
{
	static char first_call=1;
	int stat[256],stat_row[8];
	int i;
	unsigned char *ptr;
	int *i_ptr;
	if ( !do_recode ) return;
	if ( first_call ) {
		first_call = 0;
		memset ( stat,0,sizeof( stat ) );
		memset ( stat_row,0,sizeof( stat_row ) );
		for ( i = 0, ptr = buf; i < buflen; i++,ptr++ ) {
			stat[*ptr]++;
		}
		for ( i = 0, i_ptr = stat+128; i < 128; i++,i_ptr++ ) {
			stat_row[i>>4] += *i_ptr;
		}
		if ( charset == NULL ) {
			if ( (stat_row[0]+stat_row[1]+stat_row[2]+stat_row[4]+stat_row[5]+stat_row[6]+stat_row[7]) == 0 ) {
				charset = "iso-8859-1";
				do_recode = 0;
			 	return;
			} else
			if ( (stat_row[0]+stat_row[1]+stat_row[2]) >
			     (stat_row[4]+stat_row[5]+stat_row[7]) ) {
				charset = "x-cp866";
			} else if ( (stat_row[4]+stat_row[5]) == 0 ) {
				charset = "koi8-r";
			} else if ( (stat_row[6]+stat_row[7]) == 0 ) {
				charset = "x-cp1251";
			} else if ( (stat_row[4]+stat_row[5]) > (stat_row[6]+stat_row[7]) ) {
				charset = "koi8-r";
			} else {
				if ( (stat_row[0]+stat_row[1]) <
				     (stat_row[4]+stat_row[5]) ) {
					charset = "x-cp1251";
				} else {
					charset = "x-maccyr";
				}
			}
			m_log( L_INFO,"detected charset %s",charset );
		} else {
			m_log( L_INFO,"server charset %s",charset );
		}
		if ( client_charset == NULL || strcmp( charset,client_charset ) == 0 ) {
			do_recode = 0;
		}
		if ( !do_recode ) return;
		m_log ( L_INFO,"%s -> %s",charset,client_charset );
		LoadCodeTable ( code_table,charset,client_charset );
	}
	code_buf( code_table,buf,buflen );
}

#define BUF_SIZE	4096

char io_buf[BUF_SIZE];

int DumpMessage ( char *u_charset, struct Header *header, int to, int from )
{
	int nbyte=0;
	unsigned int length=0;
	char *c_length;
	ResetRecode();
	if ( header ) {
		unsigned int first_leng,total_leng;
		InitRecode ( header,u_charset );
		dump_header ( to,header );
		if ( (c_length = look_header( header,"Content-length" )) ) {
			length = atoi ( c_length );
		} else {
			if ( look_header( header,"Content-type" ) ) {
				length = (-1);
			}
		}
		if ( from == (-1) || length == 0 ) return 0;
		first_leng = length < BUF_SIZE ? length : BUF_SIZE;
		total_leng = 0;
		while ( first_leng ) {
			nbyte = read( from,io_buf+total_leng,first_leng );
			if ( nbyte == (-1) ) {
				m_perror( "read" );
				return 1;
			}
			if ( nbyte > first_leng ) {
				m_log( L_ERR,"read greater than buffer!" );
				return 1;
			}
			if ( nbyte == 0 ) break;
			total_leng += nbyte; first_leng -= nbyte;
		}
		RecodeBuffer ( io_buf,total_leng );
		write ( to,io_buf,total_leng );
		length -= total_leng;
	} else {
		length = (-1);
	}
	if ( from == (-1) ) return 0;
	while ( length && (nbyte = read( from,io_buf,length > BUF_SIZE ? BUF_SIZE : length )) > 0 ) {
		RecodeBuffer ( io_buf,nbyte );
		if ( write( to,io_buf,nbyte ) == (-1) ) {
			m_perror( "write" );
			exit( 1 );
		}
		if ( length >= nbyte ) {
			length -= nbyte;
		} else {
			length = 0;
		}
	}
	if ( nbyte == (-1) ) {
		m_perror( "read" );
		exit( 1 );
	}
	return 0;
}

void PutErrMsg ( int fd,int code,char *err,char *msg )
{
	struct Answer e_ans;
	struct Header *e_header=0;
	char *server;
	int fd_msg=(-1);
	if ( sim_prox ) return;
	server = c_malloc( sizeof(PROG_NAME)+sizeof(VERSION)+2 );
	strcpy( server,PROG_NAME );
	strcat( server,"/" );
	strcat( server,VERSION );
	e_ans.version.major = 1;
	e_ans.version.minor = 0;
	e_ans.ret_code = code;
	e_ans.answer = err;
	add_header( &e_header,"Date",http_date() );
	add_header( &e_header,"Server",server );
	if ( msg_dir ) {
		char *msg_file = c_malloc ( strlen(msg_dir)+1+strlen(msg)+6 );
		strcpy( msg_file,msg_dir );
		strcat( msg_file,"/" );
		strcat( msg_file,msg );
		strcat( msg_file,".html" );
		if ( (fd_msg = open( msg_file,O_RDONLY,0 )) != (-1) ) {
			char c_leng[15];
			add_header( &e_header,"Content-type","text/html" );
			snprintf( c_leng,14,"%d",lseek(fd_msg,0,SEEK_END) );
			add_header( &e_header,"Content-length",c_leng );
			lseek( fd_msg,0,SEEK_SET );
		}
	}
	add_header( &e_header,"\0","\0" );
	dump_answer ( fd,&e_ans );
	DumpMessage( char_arg,e_header,fd,fd_msg );
	if ( fd_msg != (-1) ) close ( fd_msg );
}

void simple_proxy( int clnt_fd, int serv_fd, char *clnt_cset, char *serv_cset )
{
  struct fd_set rfds;
  int nbyte;
  char fromserver_table[256],toserver_table[256];
  if ( LoadCodeTable( fromserver_table,serv_cset,clnt_cset ) ||
       LoadCodeTable( toserver_table,clnt_cset,serv_cset ) ) {
	m_log( L_ERR,"can not load code tables" );
	exit( 1 );
  }
  while ( 1 ) {
    int select_rc;
    FD_ZERO(&rfds);
    FD_SET(clnt_fd,&rfds);
    FD_SET(serv_fd,&rfds);
    if ( ( select_rc = select ( FD_SETSIZE,&rfds,NULL,NULL,NULL ) ) > 0 ) {
      if ( FD_ISSET(clnt_fd,&rfds) ) {
	nbyte = read ( clnt_fd,io_buf,BUF_SIZE );
	switch ( nbyte ) {
	case (-1) :
	  m_perror ( "read from client" );
	  exit( 1 );
	case 0:
	  exit( 0 );
	default:
	  code_buf ( toserver_table,io_buf,nbyte );
	  if ( write ( serv_fd,io_buf,nbyte ) == (-1) ) {
		m_perror( "write to server" );
		exit( 1 );
	  }
	  break;
	}
      }
      if ( FD_ISSET(serv_fd,&rfds) ) {
	nbyte = read ( serv_fd,io_buf,BUF_SIZE );
	switch ( nbyte ) {
	case (-1) :
	  m_perror ( "read from server" );
	  exit( 1 );
	case 0:
	  exit( 0 );
	default:
	  code_buf ( fromserver_table,io_buf,nbyte );
	  if ( write ( clnt_fd,io_buf,nbyte ) == (-1) ) {
	    m_perror( "write to client" );
	    exit( 1 );
	  }
	  break;
	}
      }
    } else if ( select_rc == 0 ) {
      continue;
    } else {
      m_perror( "select" );
    }
  }
}

extern char *optarg;
extern int   optind;

void main ( int argc, char *argv[] )
{
	struct Header *header=0,*server_header=0;
	struct Method *method=0;
	struct sockaddr_in remote,local,client,server;
	int s_leng = sizeof client;
	int r_sock;
	char *r_host,*ptr;
	short r_port;
	char *conf_file=0;
	char *client_host;
	char *s_charset;
	int ch;
	openlog( PROG_NAME,isatty(fileno(stderr)) ? LOG_PERROR|LOG_PID : LOG_PID,LOG_DAEMON );
	while ( (ch = getopt( argc,argv,"C:" )) != EOF ) {
		switch( ch ) {
		case 'C' :
			conf_file = optarg;
			break;
		default:
			syslog( LOG_ERR,"invalid switch" );
			break;
		}
	}
	argc -= optind;
	argv += optind;
	if ( *argv == NULL ) {
		syslog( LOG_ERR,"charset not defined" );
		exit( 1 );
	}
	strdcpy( &char_arg,*argv );
	if ( sup_charset( &char_arg ) ) {
		syslog( LOG_ERR,"charset %s not supported",char_arg );
		exit( 1 );
	}
	if ( load_conf( conf_file ? conf_file : DEFAULT_CONF_FILE ) ) {
		syslog( LOG_ERR,"error in config file" );
		exit( 1 );
	}
	if ( sim_prox && argc < 3 ) {
		syslog( LOG_ERR,"server name not defined" );
		exit( 1 );
	}
	if ( sim_prox ) {
		strdcpy( &s_charset,argv[1] );
		if ( sup_charset( &s_charset ) ) {
			syslog( LOG_ERR,"charset %s not supported",s_charset );
			exit( 1 );
		}
	}
	if ( log_file ) {
		m_openlog( PROG_NAME,log_file );
	}
	memset( &remote,0,sizeof remote );
	memset( &local,0,sizeof local );
	memset( &client,0,sizeof client );
	if ( getpeername ( 0,(struct sockaddr*)&client,&s_leng ) == (-1) ) {
		fprintf ( stderr,"invalid stdin\n" );
		exit( 1 );
	}
	s_leng = sizeof client;
	if ( getsockname ( 0,(struct sockaddr*)&server,&s_leng ) == (-1) ) {
		fprintf ( stderr,"invalid socket\n" );
		exit( 1 );
	}
	client_host = gethname( client.sin_addr );
	m_log( L_INFO,"%s : connect at port %d from %s",c_date(),ntohs( server.sin_port ),client_host );
	if ( is_bad_host( client_host ) ) {
		m_log( L_WARN,"access denied : %s",client_host );
		PutErrMsg( 0,400,"Forbidden","bad_host" );
		exit( 0 );
	}
	if ( sim_prox ) {
		strdcpy( &r_host,argv[2] );
		goto set_connect;
	}
	method = load_method( 0 );
	if ( method == NULL ) {
		m_log( L_WARN,"invalid request from %s",client_host );
		PutErrMsg( 0,400,"Bad request","bad_request" );
		exit( 1 );
	}
	log_method( method );
	if ( def_host == NULL || method->uri[0] != '/' ) {
		if ( strncmp( method->uri,"http://",7 ) ) {
			m_log( L_WARN,"not supported URL : %s, from %s",method->uri,client_host );
			PutErrMsg( 0,404,"Not found","not_found" );
			exit( 1 );
		}
		if ( mst_prox == NULL ) {
			ptr = strchr( method->uri+7,'/' );
			if ( ptr == NULL ) {
				ptr = strchr( method->uri+7,0 );
				*ptr = '/'; ptr[1] = 0;
			}
			r_host = c_malloc( ptr-method->uri-7+1 );
			strncpy( r_host,method->uri+7,ptr-method->uri-7 );
			method->uri = ptr;
		} else {
			r_host = mst_prox;
		}
	} else {
		r_host = def_host;
	}
	set_connect:
	r_port = 80;
	if ( (ptr = strchr( r_host,':' )) ) {
		*ptr = 0; ptr++; r_port = (short)atoi( ptr );
	}
	local.sin_family = remote.sin_family = AF_INET;
	remote.sin_port = htons( r_port );
	remote.sin_addr = getaddr( r_host );
	if ( remote.sin_addr.s_addr == 0 ) {
		m_log( L_WARN,"unknown host : %s, from %s",r_host,client_host );
		PutErrMsg( 0,404,"Host unknown","host_unknown" );
		exit( 1 );
	}
	if ( !sim_prox && method->version.major >= 1 ) {
		header = load_header( 0 );
		if ( header == NULL ) {
			m_log( L_WARN,"invalid request from %s",client_host );
			PutErrMsg( 0,400,"Bad request","bad_request" );
			exit( 1 );
		}
		log_header( header );
	}
	if ( (r_sock = socket( AF_INET,SOCK_STREAM,0 )) == (-1) ) {
		m_perror( "socket" );
		exit( 1 );
	}
	if ( our_addr ) local.sin_addr = getaddr( our_addr );
	if ( bind( r_sock,(struct sockaddr*)&local,sizeof local ) == (-1) ) {
		m_perror( "bind" );
		exit( 1 );
	}
	if ( connect( r_sock,(struct sockaddr*)&remote,sizeof remote ) == (-1) ) {
		m_perror( "connect" );
		exit( 1 );
	}
	if ( sim_prox ) {
		simple_proxy( 0,r_sock,char_arg,s_charset );
		exit( 0 );
	}
	dump_method( r_sock,method );
	if ( header || method->version.major >= 1 ) {
		struct Answer *answer;
		DumpMessage ( NULL,header,r_sock,0 );
		answer = load_answer( r_sock );
		if ( answer == NULL ) {
			m_log( L_WARN,"invalid answer from server %s",r_host );
			exit( 1 );
		}
		log_answer( answer );
		server_header = load_header( r_sock );
		if ( server_header == NULL ) {
			m_log( L_WARN,"invalid header from server %s",r_host );
			PutErrMsg( 0,400,"Bad request","bad_request" );
			exit( 1 );
		}
		log_header( server_header );
		dump_answer( 0,answer );
	}
	DumpMessage ( char_arg,server_header,0,r_sock );
	exit( 0 );
}
